# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA entities refactoring or cleanup.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
import heimdall
from heimdall.util import get_node, create_node
from lxml import etree as _et


ENTITIES = 'entities'
ATTRIBUTES = 'attributes'
PROPERTIES = 'properties'
INT_MAX_VALUE = 999999  # should theorically be `sys.maxsize * 2 + 1` but heck


def update_entities(tree, delete_orphans=False):
    """
    If ``delete_orphans`` is ``True``, metadata with neither ``aid`` nor
    ``pid`` will be deleted from their containing item, as they cannot be
    linked to any entity or property. If ``False``, these metadata will be
    left untouched.
    """
    data = dict()
    data[ENTITIES] = dict()
    data[PROPERTIES] = dict()
    all_entities = set()
    ALL_ATTRIBUTES = 'all'
    for item in heimdall.getItems(tree):
        eid = item.get('eid')
        if eid is None:
            continue  # we can't do anything if item has no entity
        entity = _get_entity(tree, data, eid)
        metas = dict()  # {aid <> counter} of already encountered metadata
        for metadata in heimdall.getMetadata(item):
            # (1) retrieve attribute and property for metadata
            aid = metadata.get('aid', None)
            attribute = _get_attribute(entity, aid)
            pid = metadata.get('pid', None)
            if attribute is not None:
                if pid is None:
                    pid = attribute.get('pid', None)
                else:  # pid is not None
                    _set_attribute_pid(attribute, pid)
            else:
                # try to deduce attribute from just eid+pid
                attribute = _get_attribute_by_pid(entity, pid)
                aid = attribute['id']
            property_ = _get_property(tree, data, pid, aid)
            if property_ is not None:
                _set_attribute_pid(attribute, property_['id'])
            else:  # pid AND aid should be null here, so metadata is orphan
                if delete_orphans:
                    metadata.getparent().remove(metadata)
                    continue  # rest of the loop is moot for orphan

            assert attribute is not None
            assert property_ is not None
            # (2) infer attribute type
            value = metadata.text
            type_ = attribute.get('type', None)
            if type_ != 'text':
                attribute['type'] = _infer_type(value)
            # (2a) infer attribute length
            # TODO: kinda useful for SQL, but not really used yet
            attribute['length'] = max(attribute.get('length', 0), len(value))
            # (3) count metadata in this item to infer min and max later
            all_attributes = entity.get(ALL_ATTRIBUTES, set())
            if eid in all_entities:
                if aid not in all_attributes:
                    # there was already an item for this entity, and this item
                    # did not have this metadata ; thus metadata is optional
                    attribute['min'] = 0
            # track that we saw this attribute, and how much times
            all_attributes.add(aid)
            entity[ALL_ATTRIBUTES] = all_attributes
            count = metas.get(aid, 0)
            count += 1
            metas[aid] = count
            # NOTE: name, description or uri cannot be infered from metadata

        # track we already saw this entity (this was the first item with it)
        all_entities.add(eid)
        # (4) update attributes min and max
        for aid, count in metas.items():
            attribute = [a for a in entity[ATTRIBUTES] if a['id'] == aid][0]
            amin = int(attribute.get('min', INT_MAX_VALUE))
            amax = int(attribute.get('max', 0))
            attribute['min'] = min(amin, count)
            attribute['max'] = max(amax, count)
    # cleanup
    for entity in data[ENTITIES].values():
        del entity[ALL_ATTRIBUTES]

    # TODO (5) infer properties types form all attributes that use it

    # (6) create/update tree nodes from infered data
    update_tree(tree, data)
    return tree


def _infer_type(value):
    try:
        float(value)
        return 'number'
    except ValueError:
        pass  # not a number
    try:
        from dateutil.parser import parse
        parse(value)
        return 'datetime'
    except ModuleNotFoundError:
        pass  # dateutils missing, we can't say if it's a date
    except:  # nopep8: E722
        # I'm not explicit on error catched because dateutil.parser.parse
        # raises custom Errors that I may be unable to import
        pass  # not a date
    return 'text'


def _get_entity(tree, data, eid):
    """Retrieves the current state of an entity as an plain object.
    """
    entity = data[ENTITIES].get(eid, None)
    if entity is not None:
        return entity
    node = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
    if node is not None:
        from heimdall.connectors.json import _entity2object
        data[ENTITIES][eid] = _entity2object(node)
    else:
        data[ENTITIES][eid] = {'id': eid, }
    return data[ENTITIES][eid]


def _get_attribute(entity, aid):
    if aid is None:
        return None
    attributes = entity.get(ATTRIBUTES, list())
    attribute = [a for a in attributes if a['id'] == aid]
    assert len(attribute) < 2
    if len(attribute) > 0:
        attribute = attribute[0]
    else:
        attribute = {'id': aid, }
        attributes.append(attribute)
        entity[ATTRIBUTES] = attributes
    return attribute


def _get_attribute_by_pid(entity, pid):
    if pid is None:
        return None
    attributes = entity.get(ATTRIBUTES, list())
    attribute = [a for a in attributes if a['pid'] == pid]
    assert len(attribute) < 2
    if len(attribute) > 0:
        attribute = attribute[0]
        if 'id' not in attribute:
            attribute['id'] = pid
    else:
        attribute = {'id': pid, 'pid': pid, }
        attributes.append(attribute)
        entity[ATTRIBUTES] = attributes
    return attribute


def _set_attribute_pid(attribute, pid):
    try:
        assert pid == attribute['pid']
    except KeyError:
        attribute['pid'] = pid


def _get_property(tree, data, pid, aid=None):
    if pid is None:
        if aid is not None:
            pid = aid
        else:
            return None
    property_ = data[PROPERTIES].get(pid, None)
    if property_ is not None:
        return property_
    node = heimdall.getProperty(tree, lambda n: n.get('id') == pid)
    if node is not None:
        from heimdall.connectors.json import _property2object
        data[PROPERTIES][pid] = _property2object(node)
    else:
        data[PROPERTIES][pid] = {'id': pid, }
    return data[PROPERTIES][pid]


def update_tree(tree, data):
    """Update HERA element tree with new data
    """
    for pid, payload in data[PROPERTIES].items():
        property_ = heimdall.getProperty(tree, lambda n: n.get('id') == pid)
        if property_ is None:
            _get_container(tree, 'properties')  # creates if missing
            heimdall.createProperty(tree, **payload)
        else:
            heimdall.updateProperty(property_, **payload)

    for eid, payload in data[ENTITIES].items():
        entity = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
        if entity is None:
            _get_container(tree, 'entities')  # creates if missing
            entity = heimdall.createEntity(tree, **payload)
        else:
            heimdall.updateEntity(entity, **payload)
        for a in payload[ATTRIBUTES]:
            aid = a['id']
            attr = heimdall.getAttribute(entity, lambda n: n.get('id') == aid)
            if attr is None:
                attr = heimdall.createAttribute(entity, **a)
            else:
                heimdall.updateAttribute(attr, **a)
            try:
                length = a['length']
                prefix = heimdall.connectors.mysql.LENGTH_PREFIX
                create_node(attr, 'rule', f'{prefix}{str(length)}')
            except KeyError:
                pass  # no length


def _get_container(tree, tag):
    container = get_node(tree, tag)
    if container is None:
        container = _et.SubElement(tree, tag)
    return container



# -*- coding: utf-8 -*-
import heimdall
from ..decorators import connector_in, connector_out
from .json import is_url, _create_tree, _tree2object
from urllib.parse import urlparse
from urllib.request import urlopen
try:
    from yaml import safe_load, dump
    available = True
except ModuleNotFoundError:
    available = False


@connector_in('hera:yaml')
def getDatabase(**options):
    r"""Imports a database from one YAML file

    :param url: URL of a YAML file to read from
    :param \**options: Keyword arguments, see below.
    :return: HERA element tree
    :rtype: lxml.ElementTree

    :Keyword arguments:
        * **encoding** (``str``, default: ``utf-8``) -- ``url`` file encoding

    This function can be used to load a database from a local or remote
    YAML-formatted HERA database file.
    For example: ::

      >>> import heimdall
      >>> url = 'path/to/a/hera/database/file.yaml'
      >>> tree = heimdall.getDatabase(format='hera:json', url=url)
    """
    if not available:
        raise ModuleNotFoundError("Module 'pyyaml' required.")
    url = options['url']
    encoding = options.get('encoding', 'utf-8')
    if not is_url(url):
        with open(url, 'r') as f:
            data = safe_load(f)
    else:
        with urlopen(url) as response:
            data = safe_load(response.read().decode(encoding))
    return _create_tree(data or dict())


@connector_out('hera:yaml')
def serialize(tree, url, **options):
    r"""Serializes a HERA elements tree into a YAML file

    :param tree: HERA elements tree
    :param url: Path of the YAML output file
    :param \**options: Keyword arguments, see below.

    :Keyword arguments:
        * **style** (``str``, default: ``block``) -- YAML flow style for
          output file. Valid values are ``flow`` of ``block``.
          Any value other than ``flow`` will be interpreted as ``block``.
        * **sort** (``bool``, default: ``False``) -- If ``True``, keys will be
          sorted in output file; if ``False`` keys will be written to ``url``
          as they are found in ``tree``.
    """
    if not available:
        raise ModuleNotFoundError("Module 'pyyaml' required.")
    data = _tree2object(tree)
    # write data to file
    style = options.get('style', 'block')
    style = (style == 'flow')
    sort = options.get('sort', False)
    with open(url, 'w', encoding='utf-8') as f:
        dump(data, f, default_flow_style=style, sort_keys=sort)



# -*- coding: utf-8 -*-

"""
Provides all `CRUD operations <https://en.wikipedia.org/wiki/Create,_read,_update_and_delete>`_ to search or edit a HERA element tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""  # nopep8: E501

from pkgutil import extend_path
from .heimdall import *

from .properties import (
    getProperty, getProperties,
    createProperty, deleteProperty,
    replaceProperty, updateProperty,
    )
from .entities import (
    getEntity, getEntities,
    createEntity, deleteEntity,
    replaceEntity, updateEntity,
    )
from .attributes import (
    getAttribute, getAttributes,
    createAttribute, deleteAttribute,
    replaceAttribute, updateAttribute,
    )
from .items import (
    getItem, getItems,
    createItem, deleteItem,
    replaceItem, updateItem,
    )
from .metadata import (
    getMetadata, getValue, getValues,
    createMetadata, deleteMetadata,
    )
from .util import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
        create_nodes as _create_nodes,
    )


__path__ = extend_path(__path__, __name__)
from .heimdall import discover
discover()


__version__ = '0.9.0'
__all__ = [
    'getDatabase',
    'createDatabase',
    'serialize',

    'getProperty', 'getProperties',
    'createProperty', 'deleteProperty',
    'replaceProperty', 'updateProperty',

    'getEntity', 'getEntities',
    'createEntity', 'deleteEntity',
    'replaceEntity', 'updateEntity',
    'getAttribute', 'getAttributes',
    'createAttribute', 'deleteAttribute',
    'replaceAttribute', 'updateAttribute',

    'getItem', 'getItems',
    'createItem', 'deleteItem',
    'replaceItem', 'updateItem',

    'getMetadata', 'getValue', 'getValues',
    'createMetadata', 'deleteMetadata',

    '__copyright__', '__license__', '__version__',
    ]



# -*- coding: utf-8 -*-
from lxml import etree as _etree
from urllib.parse import urlparse
from urllib.request import urlopen
from ..decorators import connector_in, connector_out

"""
Provides connectors to HERA-formatted XML files.

This podule defines input and output connectors to databases
composed in full or in part of XML files following the HERA schema.

* :py:class:`heimdall.connectors.xml.getDatabase` is the input connector
* :py:class:`heimdall.connectors.xml.serialize` is the output connector

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""


@connector_in('hera:xml')
def getDatabase(**options):
    url = options['url']
    if not is_url(url):
        tree = _etree.parse(url)
        # can raise OSError (file not found, ...)
    else:
        with urlopen(url) as response:
            tree = _etree.fromstring(response.read().decode())
            # can raise urllib.error.HTTPError (HTTP Error 404: Not Found, ...)
    return tree


def is_url(path):
    schemes = ('http', 'https', )
    return urlparse(path).scheme in schemes


def createDatabase():
    xml_schema = 'http://www.w3.org/2001/XMLSchema-instance'
    hera_xsd = 'https://gitlab.huma-num.fr/datasphere/hera/schema/schema.xsd'
    qname = _etree.QName(xml_schema, 'schemaLocation')
    root = _etree.Element('hera', {qname: hera_xsd})
    properties = _etree.SubElement(root, 'properties')
    entities = _etree.SubElement(root, 'entities')
    items = _etree.SubElement(root, 'items')
    return root


@connector_out('hera:xml')
def serialize(tree, url, **options):
    r"""Serialize a HERA elements tree into an XML file

    :param tree: HERA elements tree
    :param url: Path of the XML file to create
    """
    raise NotImplementedError("TODO")


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getDatabase',
    'createDatabase',
    'serialize',

    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA elements tree refactoring or cleanup.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
from lxml import etree as _et


def get_nodes(tree, tag, filter=None):
    nodes = tree.findall(f'.//{tag}')
    if filter:
        return [node for node in nodes if filter(node)]
    return nodes


def get_node(tree, tag, filter=None):
    nodes = get_nodes(tree, tag, filter)
    if len(nodes) == 0:
        return None
    if len(nodes) == 1:
        return nodes[0]
    raise IndexError(f"Too many {tag} elements ({len(nodes)})")


def get_root(tree):
    return tree.xpath('//hera')[0]


def get_language(node):
    qname = _et.QName('http://www.w3.org/XML/1998/namespace', 'lang')
    return node.get(qname)


def set_language(node, language):
    qname = _et.QName('http://www.w3.org/XML/1998/namespace', 'lang')
    return node.set(qname, language)


def create_nodes(parent, tag, text):
    nodes = list()
    if type(text) is str:
        node = create_node(parent, tag, text)
        nodes.append(node)
    else:
        for language_key, value in text.items():
            node = create_node(parent, tag, value)
            set_language(node, language_key)
            nodes.append(node)
    return nodes


def create_node(parent, tag, text=None):
    node = _et.SubElement(parent, tag)
    if text is not None:
        node.text = text
    return node


def update_node_value(node, key, value):
    if value is not None:
        node.set(key, str(value))
    else:  # remove value
        node.attrib.pop(key)


def maybe_update_node_values(node, keys, **kwargs):
    for key in keys:
        try:
            value = kwargs[key]
            update_node_value(node, key, value)
        except KeyError:
            pass  # nothing to change


def update_node_children(node, key, value):
    children = get_nodes(node, key)
    for child in children:
        if value is not None:
            child.text = str(value)
        else:  # remove child node
            child.getparent().remove(child)


def maybe_update_node_children(node, keys, **kwargs):
    for key in keys:
        try:
            value = kwargs[key]
            update_node_children(node, key, value)
        except KeyError:
            pass  # nothing to change


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit entities in a HERA element tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from lxml import etree as _et
from .util import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
        create_node as _create_node,
        create_nodes as _create_nodes,
        maybe_update_node_values as _maybe_update_values,
        maybe_update_node_children as _maybe_update_children,
    )
from .attributes import getAttribute, createAttribute, updateAttribute


def getEntity(tree, filter):
    """Retrieves a single entity.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: Entity element
    :rtype: :py:class:`lxml.etree._Element`

    This function returns a single entity from ``tree`` as defined by ``filter``, or ``None`` if there is no entity in ``tree`` corresponing to ``filter`` (or if ``tree`` has no entity).
    It works exactly like :py:class:`heimdall.getEntities`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    return _get_node(tree, 'entity', filter)


def getEntities(tree, filter=None):
    """Retrieves a collection of entities.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: List of Entity elements
    :rtype: :py:class:`list`

    This function can be used to retrieve all entities in a database: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> entities = heimdall.getEntities(tree)  # retrieve all entities

    To retrieve only *some* entities, you can use a filter.
    A filter is a function which takes only an item as a parameter, and returns either ``True`` (we want this entity to be part of the list returned by ``getEntities``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>>
      >>> def by_attribute_property(attribute):  # attribute filter
      >>>     return attribute.get('pid', 'dc:title')
      >>>
      >>> def by_property(entity):  # entity filter
      >>>     attribute = getAttribute(entity, by_attribute_property)
      >>>     return attribute is not None
      >>>
      >>> # retrieve only entities reusing a property of id 'dc:title'
      >>> entities = heimdall.getEntities(tree, by_property)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # retrieve one entity of a specific id
      >>> heimdall.getEntities(tree, lambda e: e.getValue('id') == '42')
    """  # nopep8: E501
    return _get_nodes(tree, 'entity', filter)


def createEntity(tree, id, **kwargs):
    r"""Creates a single entity.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param id: (:py:class:`str`) Entity unique identifier
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable name
        * **type** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable documentation
        * **uri** (:py:class:`list`, optional) -- URI list
    :return: Entity that was just created
    :rtype: :py:class:`lxml.etree._Element`

    This function can be used to add a new entity to a database.
    Entities created by ``createEntity`` will always be added to the ``<entities/>`` container element.

    In its simplest usage, ``createEntity`` simply creates a new ``<entity id='xxx'/>`` element: ::

      >>> import heimdall
      >>> ...
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> heimdall.createEntity(tree, 'xxx')  # create a new entity
      >>> # the following child is now added to entities list:
      >>> # <entity pid='xxx' />

    Additional supported parameters are ``type``, ``name``, ``description``, and ``uri``.
    Each of these parameters creates appropriate children for the attribute.
    Here is an example: ::

      >>> import heimdall
      >>> ...  # load HERA tree
      >>> heimdall.createEntity(tree,
      >>>     id='person', name='Person',
      >>>     description="Real person or fictional character.",
      >>>     )
      >>> # the following entity is now added to entities list:
      >>> # <entity id='person'>
      >>> #     <name>Person</name>
      >>> #     <description>Person, real or not</description>
      >>> # </entity>

    Please note that ``name`` and ``description`` can be localized, if they are
    of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows entity localization: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>>  heimdall.createEntity(tree,
      >>>      id='person',
      >>>      name={
      >>>          'de': "Person",
      >>>          'en': "Person",
      >>>          'fr': "Personne",
      >>>      },
      >>>      description={
      >>>          'en': "Person, real or not",
      >>>          'fr': "Personne réelle ou non",
      >>>      })
      >>> # the following entity is now added to the entities list:
      >>> # <entity id='person'>
      >>> #     <name xml:lang='de'>Person</name>
      >>> #     <name xml:lang='en'>Person</name>
      >>> #     <name xml:lang='fr'>Persone</name>
      >>> #     <description xml:lang='en'>Person, real or not</description>
      >>> #     <description xml:lang='fr'>Personne réelle ou non</description>
      >>> # </entity>

    Once an entity is created, some attributes can be added to it.
    For example: ::

      >>> import heimdall
      >>> ...  # load HERA tree
      >>> e = heimdall.createEntity(tree, id='person', name='Person')
      >>> heimdall.createAttribute(e, pid='name', name="Name")
      >>> heimdall.createAttribute(e, pid='eyes', name="Eye color")
      >>> # the following entity is now added to entities list:
      >>> # <entity id='person'>
      >>> #     <name>Person</name>
      >>> #     <attribute pid='name'>
      >>> #         <name>Name</name>
      >>> #     </attribute>
      >>> #     <attribute pid='eyes'>
      >>> #         <name>Eye color</name>
      >>> #     </attribute>
      >>> # </entity>

    See :py:class:`heimdall.createAttribute` for more info about adding attributes to an entity.
    """  # nopep8: E501
    container = tree.findall('.//entities')
    if container:
        container = container[0]
        nodes = container.getchildren()
    else:  # Create <entities/> node
        container = _et.SubElement(_get_root(tree), 'entities')
        nodes = []
    # Check entity unique id (eid) does not exist
    if len([n for n in nodes if n.get('id') == id]) > 0:
        raise ValueError(f"Entity id '{id}' already exists")
    # Create entity
    node = _et.SubElement(container, 'entity', id=id)
    param = kwargs.get('name', None)
    if param is not None:
        _create_nodes(node, 'name', param)
    param = kwargs.get('description', None)
    if param is not None:
        _create_nodes(node, 'description', param)
    param = kwargs.get('uri', [])
    if type(param) is not list:
        raise TypeError(f"'uri' expected:'list', got:'{type(param).__name__}'")
    for value in param:
        _create_node(node, 'uri', value)
    return node


def replaceEntity(entity, **kwargs):
    """TODO: Not Implemented
    """
    raise ValueError("TODO: Not Implemented")


def updateEntity(entity, **kwargs):
    """TODO
    """
    VALUES = ['id', ]
    CHILDREN = ['name', 'description', 'uri', ]
    _maybe_update_values(entity, VALUES, **kwargs)
    _maybe_update_children(entity, CHILDREN, **kwargs)

    attributes = kwargs.get('attributes', [])
    for attribute_payload in attributes:
        pid = attribute_payload.get('pid', None)
        aid = attribute_payload.get('id', None)

        def by_payload(a):
            return a.get('id') == aid and a.get('pid') == pid
        attribute = getAttribute(entity, by_payload)
        if attribute is None:
            createAttribute(entity, **attribute_payload)
        else:
            updateAttribute(attribute, **attribute_payload)


def deleteEntity(tree, filter):
    """Deletes a single entity.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to delete a single entity from ``tree``.
    It  performs the entity deletion "in place"; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    This function raises an ``IndexError`` if the filtering method ``filter`` returns more than one result.
    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    This method doesn't delete any item documented by the deleted entity.
    All items referencing this entity, if any, will become invalid.
    One can either delete these items afterwards, or change their ``eid``, or recreate a new entity and update these items accordingly.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete a entity using its unique id
      >>> heimdall.deleteEntity(tree, lambda e: e.get('id') == '42')
    """  # nopep8: E501
    node = _get_node(tree, 'entity', filter)
    if node is not None:
        node.getparent().remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getEntity', 'getEntities',
    'createEntity', 'deleteEntity',
    'replaceEntity', 'updateEntity',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit items in a HERA element tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

import heimdall as _h
from lxml import etree as _et
from .util import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
    )


def getItem(tree, filter):
    """Retrieves a single item.

    :param entity: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: Item element
    :rtype: :py:class:`lxml.etree._Element`

    This function returns a single item from ``tree`` as defined by ``filter``, or ``None`` if there is no item in ``tree`` corresponing to ``filter`` (or if ``tree`` has no item).
    This function works exactly like :py:class:`heimdall.getItems`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    return _get_node(tree, 'item', filter)


def getItems(tree, filter=None):
    """Retrieves a collection of items.

    :param entity: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: Item list
    :rtype: :py:class:`list`

    This function can be used to retrieve all items in a database: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> items = heimdall.getItems(tree)  # retrieve all items

    To retrieve only *some* items, you can use a filter.
    A filter is a function which takes only an item as a parameter, and returns either ``True`` (we want this item to be part of the list returned by ``getItems``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> my_favourite_author = 'Van Damme, Jean-Claude'
      >>> def by_author(item):  # create filtering function
      >>>     authors = getValues(item, 'author')
      >>>     return my_favourite_author in authors
      >>> # retrieve only items whose author is JCVD
      >>> items = heimdall.getItems(tree, by_author)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # retrieve only items of a specific id
      >>> heimdall.getItems(tree, lambda e: e.getValue('id') == '42')
    """  # nopep8: E501
    return _get_nodes(tree, 'item', filter)


def createItem(tree, **kwargs):
    r"""Creates a single item.

    :param entity: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see description.
    :return: Item that was just created
    :rtype: :py:class:`lxml.etree._Element`

    This function can be used to add a new item to a database.
    Elements created by ``createItem`` will always be added to the ``<items/>`` container element.

    With no additional keyword arguments, ``createItem`` simply creates a new ``<item/>`` element: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> heimdall.createItem(tree)  # create an empty item

    Items can be linked to a specific entity, using this entity id ``eid``.
    The following example creates the new element ``<item eid='42' />``: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>> heimdall.createItem(tree, eid='42')  # create a new item
      >>> # the corresponding entity could be created later, like this:
      >>> # heimdall.createEntity(tree, id='42', ...)

    Please note that ``createItem`` makes no consistency check, *eg.* it does not validate that an entity identified by ``eid`` exists in ``tree``.

    Additional keyword arguments to ``createItem`` each add a metadata child element to the created item.
    The name of each keyword argument is the property identifier of the metadata, and its value is the metadata value.
    The following example creates an item element with metadata: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>> heimdall.createItem(tree, name='Chirpy', type='birb')
      >>> # the following element is now added to the ``<items/>`` container element:
      >>> # <item>
      >>> #   <metadata pid='name'>Chirpy</metadata>
      >>> #   <metadata pid='type'>birb</metadata>
      >>> # </item>
      >>> # the corresponding properties could be created later, like this:
      >>> # heimdall.createProperty(tree, id='name', ...)
      >>> # heimdall.createProperty(tree, id='type', ...)

    As stated before, ``createItem`` makes no consistency check, *eg.* it does not validate that each created metadata belongs to an existing property.

    Metada added to a new item can be localized, if their value is given as a ``dict`` instead of a ``str`` (``dict`` keys are language codes).
    Here is an example: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>>  heimdall.createItem(tree, eid='pet',
      >>>      name={'en': 'Chirpy', 'fr': 'Cui-Cui', },
      >>>      type={'en': 'birb', 'fr': 'wazo', },
      >>>      floof='yes')
      >>> # the following element is now added to the items list:
      >>> # <item eid='pet'>
      >>> #   <metadata pid='name' xml:lang='en'>Chirpy</metadata>
      >>> #   <metadata pid='name' xml:lang='fr'>Cui-Cui</metadata>
      >>> #   <metadata pid='type' xml:lang='en'>birb</metadata>
      >>> #   <metadata pid='type' xml:lang='fr'>wazo</metadata>
      >>> #   <metadata pid='floof'>yes</metadata>
      >>> # </item>

    Let's state it one last time: ``createItem`` makes no consistency check.
    Thus, language codes are not verified.

    Please note that, as the ``eid`` parameter is used to create a link between the item to create and its governing entity, ``createItem`` cannot be used to create an item containing a metadata child linked to a property identified by the property id ``eid``.
    """  # nopep8: E501
    container = tree.findall('.//items')
    if container:
        container = container[0]
    else:  # Create <items/> node
        container = _et.SubElement(_get_root(tree), 'items')
    # Create item
    node = _et.SubElement(container, 'item')
    for key, value in kwargs.items():
        if key != 'eid':
            _h.createMetadata(node, key, None, value)
    param = kwargs.get('eid', None)
    if param:
        node.set('eid', param)
    return node


def replaceItem(item, **kwargs):
    """TODO: Not Implemented
    """
    raise ValueError("TODO: Not Implemented")


def updateItem(item, **kwargs):
    """TODO: Not Implemented
    """
    raise ValueError("TODO: Not Implemented")


def deleteItem(tree, filter):
    """Deletes a single item.

    :param entity: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to delete a single item from ``tree``.
    It performs the item deletion "in place"; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    This function raises an ``IndexError`` if the filtering method ``filter`` returns more than one result.
    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete an item using its unique id metadata
      >>> heimdall.deleteItem(tree, lambda e: e.getValue('id') == '42')
    """  # nopep8: E501
    node = _get_node(tree, 'item', filter)
    if node is not None:
        node.getparent().remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getItem', 'getItems',
    'createItem', 'deleteItem',
    'replaceItem', 'updateItem',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit metadata in a HERA item.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from .util import (
        get_node as _get_node,
        create_nodes as _create_nodes,
    )


def getMetadata(item, pid=None, aid=None):
    """Retrieves metadata elements from an item.

    :param item: HERA item
    :param pid: HERA property identifier
    :return: Metadata as HERA elements
    :rtype: list

    This function works exactly like ``heimdall.getValues``, but returns
    HERA metadata elements instead of `str` values.

    Here is an example: ::

      >>> import heimdall
      >>> ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item eid='person'>
      >>> #   <metadata pid='dc:name' xml:lang='en'>William</metadata>
      >>> #   <metadata pid='dc:name' xml:lang='fr'>Guillaume</metadata>
      >>> #   <metadata pid='pet'>Chirpy</metadata>
      >>> #   <metadata pid='pet'>Claws</metadata>
      >>> #   <metadata pid='pet'>Blackie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> values = heimdall.getMetadata(item, 'dc:name')
      >>> # len(values) == 2
      >>> values = heimdall.getMetadata(item, 'pet')
      >>> # len(values) == 3
      >>> values = heimdall.getMetadata(item, 'meats')
      >>> # len(values) == 0

    If ``pid`` parameter is left undefined, ``getMetadata``
    simply returns all metadata elements of an item.
    In the above example, ``heimdall.getMetadata(item)`` would
    return a list of all (5) metadata elements of ``item``.
    """
    if aid:
        xpath = './/metadata[@aid="{aid}"]'.format(aid=aid)
        return item.findall(xpath)
    if pid:
        xpath = './/metadata[@pid="{pid}"]'.format(pid=pid)
        return item.findall(xpath)
    return item.findall('.//metadata')


def getValues(item, pid):
    """Retrieves metadata values from an item.

    :param item: HERA item
    :param pid: HERA property identifier
    :return: Metadata as ``str`` values
    :rtype: ``list``

    This function can be used to retrieve metadata values from an item.
    Here is an example: ::

      >>> import heimdall
      >>> ...
      >>> item = getItem(...)
      >>> # consider we just got the following item:
      >>> # <item eid='person'>
      >>> #   <metadata pid='dc:name' xml:lang='en'>William</metadata>
      >>> #   <metadata pid='dc:name' xml:lang='fr'>Guillaume</metadata>
      >>> #   <metadata pid='pet'>Chirpy</metadata>
      >>> #   <metadata pid='pet'>Claws</metadata>
      >>> #   <metadata pid='pet'>Blackie</metadata>
      >>> # </item>
      >>> # then we'll have the following results
      >>> values = heimdall.getValues(item, 'dc:name')
      >>> # values = ['William', 'Guillaume']
      >>> values = heimdall.getValues(item, 'pet')
      >>> # values = ['Chirpy', 'Claws', 'Blackie']
      >>> values = heimdall.getValues(item, 'meats')
      >>> # values = []
    """
    xpath = './/metadata[@pid="{pid}"]'.format(pid=pid)
    values = []
    for metadata in item.findall(xpath):
        if metadata.text is not None:
            values.append(metadata.text)
        else:  # <metadata ...></metadata>
            values.append('')
    return values


def getValue(item, pid):
    """Retrieves a single metadata value from an item.

    This function works exactly like ``heimdall.getValues``, but raises an
    ``IndexError`` if there is more than one metadata corresponding to ``pid``.

    :param item: HERA item
    :param pid: HERA property identifier
    :return: Metadata value
    :rtype: ``str``
    """
    values = getValues(item, pid)
    if len(values) == 0:
        return None
    if len(values) == 1:
        return values[0]
    raise IndexError("Too many metadata ({count})".format(count=len(values)))


def createMetadata(item, pid, aid, value):
    """Adds a single metadata to an item.

    :param item: HERA item
    :param pid: HERA property identifier
    :param value: Metadata value

    Metadata created by ``createMetadata`` will always be added
    to the item ``item``, with no consistency check.
    For example, ``createMetadata`` does not verify that ``pid`` is a valid
    property identifier in the database ``item`` comes from.
    Should ``item`` have a related entity, ``createMetadata`` does not check
    that the new metadata makes sense for the entity (*ie.* that the
    entity defines an attribute reusing the property ``pid``).

    The following example adds a new metadata to an existing item: ::

      >>> import heimdall
      >>> ...
      >>> item = getItem(...)
      >>> heimdall.createMetadata(item, 'dc:name', 'name_attr', 'Bill')

    Metada added to a ``item`` can be localized, if their value is given as
    a ``dict`` instead of a ``str`` (``dict`` keys are language codes).
    Here is an example: ::

      >>> import heimdall
      >>> ...
      >>> item = getItem(...)
      >>> heimdall.createMetadata(item, 'dc:name', 'name_attr', {
      >>>     'en_AU': 'Bill',
      >>>     'de_DE': 'Wilhelm',
      >>>     })
    """
    nodes = _create_nodes(item, 'metadata', value)
    for node in nodes:
        if pid:
            node.set('pid', pid)
        if aid:
            node.set('aid', aid)
    return nodes


def deleteMetadata(item, pid):
    """TODO
    """
    node = _get_node(item, 'metadata', lambda m: m.get('pid') == pid)
    if node is not None:
        node.getparent().remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getMetadata', 'getValue', 'getValues',
    'createMetadata', 'deleteMetadata',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA properties refactoring or cleanup.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
import heimdall


def delete_unused_properties(tree, relational=True):
    """Delete unused properties from a HERA element tree.

    An unused property is not referenced by any attribute in the same tree
    (an attribute reuses a property via its ``pid``).
    Please note that if no attribute references a property, this property is
    deleted, even if one or more items in the tree reference this property.
    If an item metadata references an unused property,  the corresponding
    property is deleted anyway, as it has no use if the item's entity doesn't
    use the property via one of its attribute.

    | The previous paragraph description is valid for relational databases,
      but not for non-relational databases, where items directly use
      properties, and generally don't belong to any entities.
    | If your database is non-relational, a property isn't unused and
      shouldn't be deleted if one or more items reference it.
      To avoid this, set the ``relational`` parameter to ``False`` when using
      ``delete_unused_properties``.

    This function performs its modifications "in place".
    In other words, parameter ``tree`` is directly modified,
    and this function returns nothing.

    :param tree: HERA elements tree
    :param relational: (optional, default: ``True``) Set this parameter
       to ``False`` for non-relational specific behaviour (see description)

    Usage ::

      >>> from heimdall.util import delete_unused_properties
      >>> ...  # create config, load HERA tree
      >>> delete_unused_properties(tree)  # get rid of some clutter
    """

    # give ourselves a map of unused properties, initialized with all of them
    properties = {}
    for p in heimdall.getProperties(tree):
        properties[p.get('id')] = p
    # let's check which properties are really unused
    for e in heimdall.getEntities(tree):
        for a in heimdall.getAttributes(e):
            pid = a.get('pid')
            if pid in properties.keys():
                # this property is in use ; so we mustn't delete it
                properties.pop(pid)

    to_keep = []
    if not relational:
        for pid in properties.keys():
            for item in heimdall.getItems(tree):
                metadata = heimdall.getMetadata(item, pid)
                if len(metadata) > 0:
                    # item references pid, so property won't be deleted
                    to_keep.append(pid)

    # the map now only contains unused ones ; so, delete them
    # (if not relational, pid in `to_keep` ARE NOT deleted)
    for pid, p in properties.items():
        if pid not in to_keep:
            p.getparent().remove(p)
    # end of function, don't bother with properties.clear()


def merge_properties(tree, properties):
    """Merge duplicate properties.

    :param tree: HERA elements tree
    :param properties: Map containing old property identifiers as keys,
           and new property identifiers as values

    This function allows to merge similar properties into an existing one.
    This makes the resulting database schema more readable, because
    similarities between items and entities are more apparent when
    properties are systematically reused.

    This function updates the ``pid`` referenced by every metadata, as long as
    the ``pid`` referenced by every attribute, if these ``pid`` correspond to
    keys of the ``properties`` map parameter.
    The updated value is the key's value in ``properties``.

    Please note that only each relevant attribute ``pid`` is modified, so
    each one keeps its custom names, descriptions and whatnot.

    | As each key of ``properties`` has its own value, this method
      can be used to merge many "duplicate" properties into different
      "factorized" ones, all at once.
    | However, each value of ``properties`` must be the unique
      identifier of an existing property in ``tree``.

    After using ``merge_properties``, previous duplicate properties
    remain in place, albeit now unused.
    Thus, :py:class:`heimdall.util.delete_unused_properties` can be called
    on the same ``tree`` to get rid of them.

    Please note that this function performs its modifications "in place".
    In other words, parameter ``tree`` is directly modified,
    and this function returns nothing.

    Please also note that this function does *not* create any property.
    Thus, if some values in ``properties`` don't relate to an existing
    property identifier, ``tree`` will become inconsistent.
    To solve the problem, :py:class:`heimdall.createProperty` can be used.

    The example below shows how to reuse what is in fact the
    ``title`` property from Dublin Core, instead of entity-specific
    property names which are conceptually the same: ::

      >>> import heimdall
      >>> from heimdall.util import *
      >>> ...  # create config, load HERA tree
      >>> heimdall.createProperty(tree, 'dc:title', name="Name")
      >>> merge_properties(tree, {  # make reusage more apparent
      >>>     'book_title': 'dc:title',
      >>>     'author_name': 'dc:title',
      >>>     'character_name': 'dc:title',
      >>>     'thesaurus_keyword': 'dc:title',
      >>>     })
      >>> delete_unused_properties(tree)  # optional, but nice

    Of course, entity-specific labels can still exist: that's what attributes
    are for ; see for example functions :py:class:`heimdall.createAttribute`
    or :py:class:`heimdall.updateAttribute` for more info.
    """
    for item in heimdall.getItems(tree):
        for old, now in properties.items():
            metadata = heimdall.getMetadata(item, pid=old)
            for m in metadata:
                m.set('pid', now)

    for entity in heimdall.getEntities(tree):
        for old, now in properties.items():
            for attribute in heimdall.getAttributes(
                        entity, lambda a: a.get('pid') == old):
                heimdall.updateAttribute(attribute, pid=now)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search in or edit a HERA elements tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

CONNECTORS_IN = dict()
CONNECTORS_OUT = dict()


def discover():
    from pkgutil import iter_modules
    from importlib import import_module
    from heimdall import connectors
    for submodule in iter_modules(connectors.__path__):
        path = f'heimdall.connectors.{submodule.name}'
        import_module(path)


def getDatabase(**options):
    r"""Imports a database as a HERA element tree

    :param \**options: Keyword arguments, see below.
    :return: HERA element tree
    :rtype: lxml.ElementTree

    :Keyword arguments:
        * **url** (``str``) -- Location of the database to load
        * **format** (``str``) -- Format of the database to load, see below

    This function can be used to import an HERA element tree from different
    formats, depending of the ``format`` option.
    Supported formats are:

    * ``hera:xml``: XML file; see
      :py:class:`heimdall.connectors.xml.getDatabase`
    * ``hera:yaml``: YAML file; see
      :py:class:`heimdall.connectors.yaml.getDatabase`
    * ``hera:json``: JSON file; see
      :py:class:`heimdall.connectors.json.getDatabase`
    * ``csv``: CSV files; see
      :py:class:`heimdall.connectors.csv.getDatabase`
    * ``sql:mariadb``: MariaDB database; see
      :py:class:`heimdall.connectors.mysql.getDatabase`
    * ``sql:mysql``: MySQL database; see
      :py:class:`heimdall.connectors.mysql.getDatabase`

    Depending on ``format`` option, ``getDatabase`` may accept more options.
    See the individual module ``getDatabase`` documentation for more info.
    """
    fun = CONNECTORS_IN[options['format']]
    return fun(**options)


def createDatabase():
    """Creates an empty database

    :return: HERA element tree
    :rtype: lxml.ElementTree
    """
    from .connectors import xml
    return xml.createDatabase()


def serialize(tree, **options):
    r"""Exports a HERA element tree

    :param tree: HERA element tree
    :param \**options: (optional) Keyword arguments, see description.

    This function can be used to export an HERA element tree in different
    formats, depending of the ``format`` parameter.
    Supported formats are:

    * ``hera:xml``: XML file; see
      :py:class:`heimdall.connectirs.xml.serialize`
    * ``hera:yaml``: YAML file; see
      :py:class:`heimdall.connectors.yaml.serialize`
    * ``hera:json``: JSON file; see
      :py:class:`heimdall.connectors.json.serialize`
    * ``csv``: CSV files; see
      :py:class:`heimdall.connectors.csv.serialize`
    * ``sql:mariadb``: MariaDB database; see
      :py:class:`heimdall.connectors.mysql.serialize`
    * ``sql:mysql``: MySQL database; see
      :py:class:`heimdall.connectors.mysql.serialize`

    Depending on ``format`` option, ``serialize`` may accept more options.
    See the individual module ``serialize`` documentation for more info.
    """
    fun = CONNECTORS_OUT[options['format']]
    return fun(tree, **options)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getDatabase',
    'createDatabase',
    'serialize',

    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit attributes in a HERA entity.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from lxml import etree as _et
from .util import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
        create_node as _create_node,
        create_nodes as _create_nodes,
        maybe_update_node_values as _maybe_update_values,
        maybe_update_node_children as _maybe_update_children,
    )


def getAttribute(entity, filter):
    """Retrieves a single attribute from an entity.

    :param entity: (:py:class:`lxml.etree._Element`) -- HERA entity
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: Attribute from ``entity``
    :rtype: :py:class:`lxml.etree._Element`

    This function works exactly like :py:class:`heimdall.getAttributes`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    return _get_node(entity, 'attribute', filter)


def getAttributes(entity, filter=None):
    """Retrieves attributes from an entity.

    :param entity: (:py:class:`lxml.etree._Element`) -- HERA entity
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: Attributes subset from ``entity``
    :rtype: :py:class:`list`

    This function can be used to retrieve all attributes of an entity: ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> attributes = heimdall.getAttributes(enitity)  # retrieve all attributes

    To retrieve only *some* attributes, you can use a filter.
    A filter is a function which takes only one attribute as a parameter, and returns either ``True`` (we want this attribute to be part of the list returned by ``getAttributes``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> def by_pid(attribute):  # create filtering function
      >>>     return attribute.get('pid') == 'dc:title'
      >>> # retrieve only attributes equivalent to a title
      >>> names = heimdall.getAttributes(entity, by_pid)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create or load HERA tree
      >>> entity = heimdall.getEntity(tree, ...)  # retrieve an entity
      >>> # retrieve only attributes equivalent to an identifier
      >>> ids = heimdall.getAttributes(entity, lambda a: a.get('pid') == 'id')
    """  # nopep8: E501
    return _get_nodes(entity, 'attribute', filter)


def createAttribute(entity, pid, **kwargs):
    r"""Adds a single attribute to an entity.

    :param entity: HERA entity
    :param pid: (:py:class:`str`) Linked property ID
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **id** (:py:class:`str`, optional) -- Attribute id
        * **min** (:py:class:`int`, optional, default: ``0``) -- Minimum occurences in an item: ``1`` or more means it is mandatory
        * **max** (:py:class:`int`, optional) -- Maximum occurences in an item: ``1`` means it is not repeatable, more (or no value) means it is; ``0`` or ``<min`` is undefined behaviour
        * **type** (:py:class:`str`, optional) -- Type override; no value means the linked property ``type`` will be used
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Name override; no value means the linked property ``name`` will be used
        * **description** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Description override; no value means the linked property ``description`` will be used
        * **uri** (:py:class:`list`, optional) -- URI list override; no value means the linked property ``uri`` will be used, any value adds to the property ``uri``
    :return: Attribute that was just created
    :rtype: :py:class:`lxml.etree._Element`

    This function can be used to add a new attribute to an existing entity.
    Attributes created by ``createAttribute`` will always be added to the
    entity ``entity``, with no consistency check.
    For example, ``createAttribute`` does not verify that ``pid`` is a valid
    property identifier in the database ``entity`` comes from.

    In its simplest usage, ``createAttribute`` simply creates
    a new ``<attribute pid='xxx'/>`` element: ::

      >>> import heimdall
      >>> ...
      >>> entity = heimdall.getEntity(...)
      >>> heimdall.createAttribute(entity, pid='xxx')  # create a new attribute
      >>> # the following child is now added to entity attributes list:
      >>> # <attribute pid='xxx' />

    Additional supported parameters are ``type``, ``name``, ``description``,
    and ``uri``.
    Each of these parameters creates appropriate children for the attribute.
    Here is an example: ::

      >>> import heimdall
      >>> ...
      >>> heimdall.createAttribute(entity,
      >>>     pid='dc:title', name='Label',
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://purl.org/dc/terms/alternative',
      >>>         'http://nakala.fr/terms#title'
      >>>     ])
      >>> # the following attribute is now added to the entity:
      >>> # <attribute pid='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name>Label</name>
      >>> #     <uri>http://purl.org/dc/terms/title</uri>
      >>> #     <uri>http://purl.org/dc/terms/alternative</uri>
      >>> #     <uri>http://nakala.fr/terms#title</uri>
      >>> # </attribute>

    Please note that ``name`` and ``description`` can be localized, if they are
    of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows attribute localization: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>>  heimdall.createAttribute(entity,
      >>>      pid='dc:title',
      >>>      max='1',
      >>>      name={
      >>>          'de': "Name",
      >>>          'en': "Name",
      >>>          'fr': "Nom",
      >>>      },
      >>>      description={
      >>>          'en': "Human-readable name",
      >>>          'fr': "Nom usuel",
      >>>      })
      >>> # the following element is now added to the corresponding ``<entity/>`` element:
      >>> # <attribute pid='dc:title' min='0' max='1'>
      >>> #     <type>text</type>
      >>> #     <name xml:lang='de'>Name</name>
      >>> #     <name xml:lang='en'>Name</name>
      >>> #     <name xml:lang='fr'>Nom</name>
      >>> #     <description xml:lang='en'>Human-readable name</description>
      >>> #     <description xml:lang='fr'>Nom usuel</description>
      >>> # </attribute>
    """  # nopep8: E501
    # TODO check pid is not already in use in the entity
    node = _et.SubElement(entity, 'attribute', pid=pid)

    param = kwargs.get('id', None)
    if param is not None:
        # Check attribute unique id (aid) does not exist
        if getAttribute(entity, lambda a: a.get('id') == param) is not None:
            raise ValueError(f"Attribute id '{param}' already exists")
        node.set('id', str(param))
    param = kwargs.get('min', 0)
    node.set('min', str(int(param)))
    param = kwargs.get('max', None)
    if param is not None:
        node.set('max', str(int(param)))

    param = kwargs.get('type', None)
    if param is not None:
        _create_node(node, 'type', param)
    param = kwargs.get('name', None)
    if param is not None:
        _create_nodes(node, 'name', param)
    param = kwargs.get('description', None)
    if param is not None:
        _create_nodes(node, 'description', param)
    param = kwargs.get('uri', [])
    if type(param) is not list:
        raise TypeError(f"'uri' expected:'list', got:'{type(param).__name__}'")
    for value in param:
        _create_node(node, 'uri', value)
    return node


def replaceAttribute(attribute, **kwargs):
    """TODO: Not Implemented
    """
    raise ValueError("TODO: Not Implemented")


def updateAttribute(attribute, **kwargs):
    """TODO
    """
    VALUES = ['id', 'pid', 'min', 'max', ]
    CHILDREN = ['type', 'name', 'description', 'uri', ]
    _maybe_update_values(attribute, VALUES, **kwargs)
    _maybe_update_children(attribute, CHILDREN, **kwargs)


def deleteAttribute(entity, filter):
    """Deletes a single Attribute from an entity.

    :param entity: (:py:class:`lxml.etree._Element`) -- HERA entity
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to remove an attribute from an entity.
    It performs the attribute deletion "in place" ; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    This function raises an ``IndexError`` if the filtering method ``filter`` returns more than one result.
    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    This function doesn't delete any metadata referencing the ``pid`` of the deleted attribute.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete an attribute of a specific entity
      >>> e = heimdall.getEntity(tree, lambda e : e.get('id') == 'person')
      >>> heimdall.deleteAttribute(e, lambda a: a.get('pid') == 'religion')
    """  # nopep8: E501
    node = _get_node(entity, 'attribute', filter)
    if node is not None:
        node.getparent().remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getAttribute', 'getAttributes',
    'createAttribute', 'deleteAttribute',
    'replaceAttribute', 'updateAttribute',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-

"""
Provide connectors to various types of databases.

In addition to the connectors available here as submodules, anyone (for example, *you*!) can implement her or his own connector for any given database format, with no need to integrate them in pyHeimdall proper whatsoever.
Any "third party" (from pyHeimdall contributors point of view) connector will be automatically detected and made available by pyHeimdall, with no need to import the connector explicitely.
This connector will be usable without any need for the user to know which connector supports which format, simply by using :py:class:`heimdall.getDatabase` or :py:class:`heimdall.serialize` as usual.
This is true as long as both pyHeimdall and the connector are available (*ie.* installed) on the computer.

Examples of custom connectors independent from pyHeimdall proper are made available `in the following repositories <https://gitlab.huma-num.fr/datasphere/heimdall/connectors>`_.
For the interested reader, `this example connector <https://gitlab.huma-num.fr/datasphere/heimdall/connectors/example/-/blob/main/src/heimdall/connectors/example.py>`_ can be used as a tutorial, for example by following the following steps:

#. Fork `the example repository <https://gitlab.huma-num.fr/datasphere/heimdall/connectors/example/>`_
#. Rename ``src/heimdall/connectors/example.py`` as something you like.
#. Modify it to implement your new behaviour(s).
#. Don't forget to change the format your connector will be registered as in the parameter of ``@connector_in`` and/or ``@connector_out``.

| ... aaand you're done!
| Don't forget to publish it, share it, and welcome in the community!

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""  # nopep8: E501

from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)



# -*- coding: utf-8 -*-
import heimdall
from heimdall.util import get_node, get_nodes, get_language
from ..decorators import connector_in, connector_out
from json import load, loads, dump
from urllib.parse import urlparse
from urllib.request import urlopen
from lxml import etree


@connector_in('hera:json')
def getDatabase(**options):
    r"""Imports a database from one JSON file

    :param url: URL of a JSON file to read from
    :param \**options: Keyword arguments, see below.
    :return: HERA element tree
    :rtype: lxml.ElementTree

    :Keyword arguments:
        * **encoding** (``str``, default: ``utf-8``) -- ``url`` file encoding

    This function can be used to load a database from a local or remote
    JSON-formatted HERA database file.
    For example: ::

      >>> import heimdall
      >>> url = 'path/to/a/hera/database/file.json'
      >>> tree = heimdall.getDatabase(format='hera:json', url=url)
    """
    url = options['url']
    encoding = options.get('encoding', 'utf-8')
    if not is_url(url):
        with open(url, 'r') as f:
            data = load(f)
    else:
        with urlopen(url) as response:
            data = loads(response.read().decode(encoding))
    return _create_tree(data)


def is_url(path):
    schemes = ('http', 'https', )
    return urlparse(path).scheme in schemes


def _create_tree(data):
    root = heimdall.createDatabase()

    # create Properties if any
    properties = data.get('properties', None)
    if properties is not None:
        elements = root.xpath('//properties')[0]
        for o in properties:
            uid = o['id']  # id is mandatory
            e = etree.SubElement(elements, 'property', id=uid)
            _add_xml_element_from_json(e, o, 'type')
            _add_xml_element_from_json(e, o, 'name')
            _add_xml_element_from_json(e, o, 'description')
            _add_xml_element_from_json(e, o, 'uri')
            # _add_xml_element_from_json(e, o, 'uris')  # TODO property.uris

    # create Entities if any
    entities = data.get('entities', None)
    if entities is not None:
        elements = root.xpath('//entities')[0]
        for o in entities:
            uid = o['id']  # id is mandatory
            e = etree.SubElement(elements, 'entity', id=uid)
            _add_xml_element_from_json(e, o, 'name')
            _add_xml_element_from_json(e, o, 'description')
            _add_xml_element_from_json(e, o, 'uri')
            attributes = o.get('attributes', list())
            for c in attributes:
                pid = c['pid']  # pid is mandatory
                a = etree.SubElement(e, 'attribute', pid=pid)
                aid = c.get('id', None)
                if aid is not None:
                    a.attrib['id'] = aid
                min = c.get('min', 0)
                a.attrib['min'] = min
                max = c.get('max', None)
                if max is not None:
                    a.attrib['max'] = max
                _add_xml_element_from_json(a, c, 'type')
                _add_xml_element_from_json(a, c, 'name')
                _add_xml_element_from_json(a, c, 'description')
                _add_xml_element_from_json(a, c, 'uri')

    # create Items if any
    items = data.get('items', None)
    if items is not None:
        elements = root.xpath('//items')[0]
        for o in items:
            e = etree.SubElement(elements, 'item')
            _add_attribute_from_json(e, o, 'eid')
            metadata = o.get('metadata', [])
            for m in metadata:
                pid = m['pid']  # pid is mandatory
                meta = etree.SubElement(e, 'metadata', pid=pid)
                meta.text = m['value']  # value is mandatory, too

    return etree.ElementTree(root)


def _add_xml_element_from_json(e, o, attr):
    value = o.get(attr, None)
    if value is not None:
        sub = etree.SubElement(e, attr)
        sub.text = value


def _add_attribute_from_json(e, o, attr):
    value = o.get(attr, None)
    if value is not None:
        e.set(attr, value)


@connector_out('hera:json')
def serialize(tree, url, **options):
    r"""Serializes a HERA elements tree into a JSON file

    :param tree: HERA elements tree
    :param url: Path of the JSON file to create

    .. ERROR::
       This feature is not implemented yet.
       This can be either due to lack of resources, lack of demand, because it
       wouldn't be easily maintenable, or any combination of these factors.

       Interested readers can submit a request to further this topic.
       See the ``CONTRIBUTING`` file at the root of the repository for details.
    """
    data = _tree2object(tree)
    # write data to file
    indent = options.get('indent', 2)
    protek = options.get('ensure_ascii', False)
    with open(url, 'w', encoding='utf-8') as f:
        dump(data, f, indent=indent, ensure_ascii=protek)


def _tree2object(tree):
    data = dict()
    # convert properties to json-ready objects, add them to data
    elements = heimdall.getProperties(tree)
    if len(elements) > 0:
        data['properties'] = list()
    for element in elements:
        data['properties'].append(_property2object(element))
    # convert entities to json-ready objects, add them to data
    elements = heimdall.getEntities(tree)
    if len(elements) > 0:
        data['entities'] = list()
    for element in elements:
        data['entities'].append(_entity2object(element))
    # convert items to json-ready objects, add them to data
    elements = heimdall.getItems(tree)
    if len(elements) > 0:
        data['items'] = list()
    for element in elements:
        data['items'].append(_item2object(element))

    return data


def _property2object(p):
    return _xml2object(p)


def _entity2object(e):
    data = _xml2object(e)
    elements = heimdall.getAttributes(e)
    if len(elements) > 0:
        data['attributes'] = list()
    for element in elements:
        data['attributes'].append(_attribute2object(element))
    return data


def _attribute2object(a):
    data = _xml2object(a)
    data['pid'] = a.get('pid')
    data['min'] = a.get('min', 0)
    _param2text(a, data, 'max', None)
    return data


def _xml2object(x):
    data = dict()
    value = x.get('id', None)
    if value is not None:
        data['id'] = value
    # child --> unique value
    for tag in ['type', ]:
        array_of_one = [c.text for c in x.getchildren() if c.tag == tag]
        assert len(array_of_one) < 2
        if len(array_of_one) > 0 and array_of_one[0] is not None:
            data[tag] = array_of_one[0]
    # child --> localized object
    for tag in ['name', 'description', ]:
        obj = _nodes2object(get_nodes(x, tag))
        if obj is not None:
            data[tag] = obj
    # child --> repeatable value
    for tag in ['uri', ]:
        array = _nodes2array(get_nodes(x, tag))
        if len(array) > 0:
            data[tag] = array
    return data


def _item2object(i):
    data = dict()
    for key in ['eid', ]:
        _param2text(i, data, key)
    elements = heimdall.getMetadata(i)
    if len(elements) > 0:
        data['metadata'] = list()
    for element in elements:
        data['metadata'].append(_metadata2object(element))
    return data


def _metadata2object(m):
    data = dict()
    for key in ['pid', 'aid', ]:
        _param2text(m, data, key)
    value = get_language(m)
    if value is not None:
        data['language'] = value
    value = _node2text(m)
    if value is not None:
        data['value'] = value
    return data


def _param2text(x, data, key, default=None):
    value = x.get(key, default)
    if value is not None:
        data[key] = value


def _node2text(node):
    return node.text


def _nodes2object(nodes):
    data = dict()
    for node in nodes:
        language = get_language(node)
        if language is not None:
            data[language] = node.text
        else:
            # emergency exit:
            # if there is even only 1 node with no language,
            # we'll return this one only
            return node.text
    if len(data) < 1:
        return None
    return data


def _nodes2array(nodes):
    data = list()
    for node in nodes:
        if node.text is not None:
            data.append(node.text)
    return data



# -*- coding: utf-8 -*-

"""
Provides CRUD operations to search for or
edit properties in a HERA element tree.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""

from lxml import etree as _et
from .util import (
        get_node as _get_node,
        get_nodes as _get_nodes,
        get_root as _get_root,
        create_node as _create_node,
        create_nodes as _create_nodes,
        maybe_update_node_values as _maybe_update_values,
        maybe_update_node_children as _maybe_update_children,
    )


def getProperty(tree, filter):
    """Retrieves a single property.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: Property element
    :rtype: :py:class:`lxml.etree._Element`

    This function returns a single property from ``tree`` as defined by ``filter``, or ``None`` if there is no property in ``tree`` corresponing to ``filter`` (or if ``tree`` has no property).
    It works exactly like :py:class:`heimdall.getProperties`, but raises an ``IndexError`` if ``filter`` returns more than one result.
    """  # nopep8: E501
    return _get_node(tree, 'property', filter)


def getProperties(tree, filter=None):
    """Retrieves a collection of properties.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`, optional) -- Filtering function
    :return: List of Property elements
    :rtype: :py:class:`list`

    This function can be used to retrieve all properties in a database: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> properties = heimdall.getProperties(tree)  # retrieve all properties

    To retrieve only *some* properties, you can use a filter.
    A filter is a function which takes only a property as a parameter, and returns either ``True`` (we want this property to be part of the list returned by ``getProperties``) or ``False`` (we don't want it). ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> def by_pid(property):  # create filtering function
      >>>     return property.get('id') == 'dc:title'
      >>> # retrieve only a specific property by its id
      >>> properties = heimdall.getProperties(tree, by_pid)

    For simple filters, anonymous functions can of course be used: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # retrieve one specific property by its id
      >>> heimdall.getProperties(tree, lambda e: e.get('id') == 'dc:title')
    """  # nopep8: E501
    return _get_nodes(tree, 'property', filter)


def createProperty(tree, id, **kwargs):
    r"""Creates a single property.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param id: (:py:class:`str`) Property unique identifier
    :param \**kwargs: (:py:class:`dict`, optional) -- Keyword arguments, see below.
    :Keyword arguments:
        * **type** (:py:class:`str`, optional, default: ``text``) -- Property type
        * **name** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable name
        * **type** (:py:class:`str`\ \|\ :py:class:`dict`, optional) -- Human-readable documentation
        * **uri** (:py:class:`list`, optional) -- URI list
    :return: Property that was just created
    :rtype: :py:class:`lxml.etree._Element`

    This function can be used to add a new property to a database.
    Elements created by ``createProperty`` will always be added to the ``<properties/>`` container element.

    In its simplest usage, ``createProperty`` simply creates a new ``<property id='xxx'/>`` element: ::

      >>> import heimdall
      >>> ...  # create config
      >>> tree = heimdall.getDatabase(config)  # load HERA tree
      >>> heimdall.createProperty(tree, id='xxx')  # create a new property
      >>> # the following child is now added to the properties list:
      >>> # <property id='xxx' />

    Each property identifier must be unique among all properties.
    Thus, calling ``createProperty`` with an ``id`` already in use by another property in the same database will fail.

    Additional supported parameters are ``type``, ``name``, ``description``, and ``uri``.
    Each of these parameters creates appropriate children for the property.
    Here is an example: ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> heimdall.createProperty(tree,
      >>>     id='dc:title', name='Title',
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://purl.org/dc/terms/alternative',
      >>>         'http://nakala.fr/terms#title'
      >>>     ])
      >>> # the following property is now added to the properties list:
      >>> # <property id='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name>Title</name>
      >>> #     <uri>http://purl.org/dc/terms/title</uri>
      >>> #     <uri>http://purl.org/dc/terms/alternative</uri>
      >>> #     <uri>http://nakala.fr/terms#title</uri>
      >>> # </property>

    Please note that ``name`` and ``description`` can be localized, if they are of type :py:class:`dict` instead of :py:class:`str` (:py:class:`dict` keys are language codes).
    The following example shows property localization: ::

      >>> import heimdall
      >>> ...  # create HERA element tree
      >>>  heimdall.createProperty(tree, id='dc:title',
      >>>      name={
      >>>          'de': "Titel",
      >>>          'en': "Title",
      >>>          'fr': "Titre",
      >>>      },
      >>>      description={
      >>>          'en': "A name given to the resource.",
      >>>          'fr': "Un nom donné à la ressource.",
      >>>      })
      >>> # the following XML element is now added to the ``<properties/>`` container element:
      >>> # <property id='dc:title'>
      >>> #     <type>text</type>
      >>> #     <name xml:lang='de'>Titel</name>
      >>> #     <name xml:lang='en'>Title</name>
      >>> #     <name xml:lang='fr'>Titre</name>
      >>> #     <description xml:lang='en'>A name given to the resource.</description>
      >>> #     <description xml:lang='fr'>Un nom donné à la ressource.</description>
      >>> # </property>
    """  # nopep8: E501
    container = tree.findall('.//properties')
    if container:
        container = container[0]
        nodes = container.getchildren()
    else:  # Create <properties/> node
        container = _et.SubElement(_get_root(tree), 'properties')
        nodes = []
    # Check property unique id (pid) does not exist
    if len([n for n in nodes if n.get('id') == id]) > 0:
        raise ValueError(f"Property id '{id}' already exists")
    # Create property
    node = _et.SubElement(container, 'property', id=id)
    param = kwargs.get('type', 'text')
    _create_node(node, 'type', param)
    param = kwargs.get('name', None)
    if param:
        _create_nodes(node, 'name', param)
    param = kwargs.get('description', None)
    if param:
        _create_nodes(node, 'description', param)
    param = kwargs.get('uri', [])
    if type(param) is not list:
        raise TypeError(f"'uri' expected:'list', got:'{type(param).__name__}'")
    for value in param:
        _create_node(node, 'uri', value)
    return node


def replaceProperty(property, **kwargs):
    """TODO: Not Implemented
    """
    raise ValueError("TODO: Not Implemented")


def updateProperty(property, **kwargs):
    """TODO
    """
    VALUES = ['id', ]
    CHILDREN = ['type', 'name', 'description', 'uri', ]
    _maybe_update_values(property, VALUES, **kwargs)
    _maybe_update_children(property, CHILDREN, **kwargs)


def deleteProperty(tree, filter):
    """Deletes a single property.

    :param tree: (:py:class:`lxml.etree._ElementTree`) -- HERA element tree
    :param filter: (:py:class:`function`) -- Filtering function
    :return: None
    :rtype: :py:class:`NoneType`

    This function can be used to delete a single property from ``tree``.
    It performs the property deletion "in place"; in other words, parameter ``tree`` is directly modified, and this function returns nothing.

    This function raises an ``IndexError`` if the filtering method ``filter`` returns more than one result.
    If ``filter`` returns no result, this function does nothing, and does not raise any error.

    This method doesn't delete any metadata, nor any attribute, referencing the deleted property.
    All elements referencing this property, if any, will become invalid.
    They should either be deleted or fixed accordingly.

    Usage ::

      >>> import heimdall
      >>> ...  # create config, load HERA tree
      >>> # delete a property using its unique id
      >>> heimdall.deleteProperty(tree, lambda e: e.get('id') == '42')
    """  # nopep8: E501
    node = _get_node(tree, 'property', filter)
    if node is not None:
        node.getparent().remove(node)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'
__all__ = [
    'getProperty', 'getProperties',
    'createProperty', 'deleteProperty',
    'replaceProperty', 'updateProperty',
    '__copyright__', '__license__',
    ]



# -*- coding: utf-8 -*-
import heimdall
import os as _os
import re as _re
from sys import version_info as _py
from urllib.parse import urlparse
from lxml import etree
from ..decorators import connector_in, connector_out
_python_version = (_py.major, _py.minor) >= (3, 8)
try:
    from mysql.connector import connect
    _installed = True
except ModuleNotFoundError:
    _installed = False
except SyntaxError:
    _python_version = False
available = _installed and _python_version
LENGTH_PREFIX = 'len:'


def check_available():
    if not _python_version:
        version = '.'.join(str(n) for n in _py[:3])
        message = f"Python 3.8 or later required (found: {version})."
        raise ModuleNotFoundError(message)
    if not _installed:
        message = "Module 'mysql-connector-python' required."
        raise ModuleNotFoundError(message)
    return available


@connector_in(['sql:mariadb', 'sql:mysql', ])
def getDatabase(**options):
    r"""Loads a MySQL database as a HERA elements tree

    :param \**options: Keyword arguments, see below.
    :Keyword arguments:
        * **url** (``str``) -- URL of the database to load
        * **entities** (``list``) -- List of tables to load;
          empty list will do nothing

    Regarding SQL to HERA transformation, the following apply:

    * Each of the SQL table name in the ``entities`` option
      will be loaded as a single HERA entity.
    * Each column of a table will be loaded as a single HERA
      attribute, referencing a single HERA property.
      This means that, for example, two tables with the same
      ``id`` primary key column will become two different
      attribues, referencing two different properties.
      These properties can then be factorized using pyHeimdall.
      See :py:class:`heimdall.util.merge_properties` module for details.
    """
    check_available()  # breaks if not
    connection = _connect(options['url'])
    with connection.cursor() as cursor:
        hera = _create_tree(options['entities'], cursor)
    connection.close()
    return hera


def _connect(url):
    url = urlparse(url)
    # due to urlparse, url.path is something like '/dbname'
    # but mysql.connector.connect wants database = 'dbname'
    connection = connect(database=url.path.split('/')[1],
                         user=url.username, password=url.password,
                         host=url.hostname, port=url.port)
    # TBD: can connection.is_connected() be False here?
    return connection


def _create_tree(tables, cursor):
    root = heimdall.createDatabase()
    properties = root.xpath('//properties')[0]
    entities = root.xpath('//entities')[0]
    items = root.xpath('//items')[0]
    for table in tables:
        # create entity for this table
        entity, aid_vs_property = _create_entity(table, cursor)
        entities.append(entity)
        # create properties for this entity
        for p in aid_vs_property.values():
            properties.append(p)
        # create items for this entity
        eid = entity.get('id')
        result = cursor.execute(f'SELECT * FROM {table}')
        for row in cursor.fetchall():
            items.append(_create_item(eid, row, aid_vs_property))
    return root


def _create_item(eid, row, aid_vs_property):
    item = etree.Element('item', {'eid': eid, })
    for index, (aid, p) in enumerate(aid_vs_property.items()):
        value = row[index]
        if value is None:
            continue
        item.append(_create_metadata(value, aid=aid, pid=p.get('id')))
    return item


def _create_metadata(value, pid=None, aid=None):
    metadata = etree.Element('metadata')
    if aid is not None:
        metadata.set('aid', aid)
    if pid is not None:
        metadata.set('pid', pid)
    metadata.text = str(value)
    return metadata


def _create_entity(table, cursor):
    cursor.execute(f'SHOW CREATE TABLE {table}')
    create_table_query = cursor.fetchall()[0][1]
    entity = etree.Element('entity', {'id': table, })
    etree.SubElement(entity, 'name').text = table
    comment = _get_table_comment(create_table_query)
    if comment is not None:
        etree.SubElement(entity, 'description').text = comment
    pointers = _get_pointers(table, cursor)
    aid_vs_property = dict()
    cursor.execute(f'SHOW FULL COLUMNS FROM {table}')
    for row in cursor.fetchall():
        a = _create_attribute(row, table, pointers)
        entity.append(a)
        aid_vs_property[a.get('id')] = _create_property(table, row)
    return entity, aid_vs_property


def _get_table_comment(create_table_query):
    pattern = _re.compile(r"COMMENT='(?P<res>[\w\s]*)'")
    m = pattern.search(create_table_query)
    return m.group('res') if m is not None else None


def _get_pointers(source_table, cursor):
    """Gets pointer attributes for ``source_table``

    Each foreign key in ``source_table`` in SQL is a pointer in HERA.
    For example if ``SHOW CREATE TABLE source_table`` returns something such as
    ``FOREIGN KEY (`source_attr`) REFERENCES `target_table` (`target_attr`)``,
    then the attribute ``source_attr`` of entity ``source_table`` is a pointer
    to the content of attribute ``target_attr`` of entity ``target_table``.
    """
    cursor.execute(f'SHOW CREATE TABLE {source_table}')
    parts = cursor.fetchall()
    assert len(parts) == 1
    assert len(parts[0]) == 2
    assert parts[0][0] == source_table
    haystack = parts[0][1]
    needle = _re.compile(r"FOREIGN KEY \(`(?P<source_attr>\w+)`\) REFERENCES `(?P<target_table>\w+)` \(`(?P<target_attr>\w+)`\)", _re.IGNORECASE)  # nopep8: E501
    groups = needle.findall(haystack)
    pointers = dict()
    for match in groups:
        source_attr = f'{match[0]}_attr'
        target_table = match[1]
        target_attr = f'{match[2]}_attr'
        pointers[source_attr] = f'@{target_table}.{target_attr}'
    return pointers


def _create_attribute(row, table, pointers=None):
    # @see https://dev.mysql.com/doc/refman/8.4/en/show-columns.html
    name = row[0]
    sqltype = row[1]
    collation = row[2]
    nullability = row[3]  # YES|NO
    indexed = row[4]  # PRI|UNI|MUL
    default_value = row[5]
    extra = row[6]
    privileges = row[7]
    comment = row[8]
    root = etree.Element('attribute', {
        'id': f'{table}.{name}_attr', 'pid': f'{table}.{name}',
        'min': str(0) if nullability == 'YES' else str(1),
        'max': str(1),  # TODO hera allows repeatability, sql does not (as is)
        })
    pointers = pointers or dict()
    pointer_type = pointers.get(f'{name}_attr', None)
    if pointer_type is not None:
        etree.SubElement(root, 'type').text = pointer_type
    else:
        etree.SubElement(root, 'type').text = _type_sql2hera(sqltype)
    etree.SubElement(root, 'name').text = name
    if comment:
        etree.SubElement(root, 'description').text = comment
    return root


def _create_property(table, row):
    # @see https://dev.mysql.com/doc/refman/8.4/en/show-columns.html
    name = row[0]
    sqltype = row[1]
    comment = row[8]
    root = etree.Element('property', id=f'{table}.{name}')
    etree.SubElement(root, 'type').text = _type_sql2hera(sqltype)
    etree.SubElement(root, 'name').text = name
    if comment is not None:
        etree.SubElement(root, 'description').text = comment
    return root


def _type_sql2hera(sqltype):
    # NOTE: the following lines might allow python3.7 support
    # if type(sqltype) == bytes:
    #     try:
    #         import chardet  # happens with python3.7 or lower
    #         encoding = chardet.detect(sqltype)['encoding']
    #     except ModuleNotFoundError:
    #         encoding = 'ascii'
    #     sqltype = sqltype.decode(encoding)
    if (sqltype == 'date' or
            sqltype == 'datetime' or
            sqltype == 'timestamp'):
        return 'datetime'
    if (sqltype.startswith('varchar') or
            sqltype.startswith('char') or
            sqltype.startswith('tinytext')):
        return 'text'
    if (sqltype.startswith('int') or
            sqltype.startswith('tinyint')):
        return 'number'
    raise ValueError(f"Unknown type '{sqltype}'")


@connector_out(['sql:mariadb', 'sql:mysql', ])
def serialize(tree, url, **options):
    r"""Serialize a HERA elements tree into a MySQL dump like file

    :param tree: HERA elements tree
    :param url: Path of the MySQL dump file to create
    :param \**options: (optional) Keyword arguments, see below.
    :Keyword arguments:
        * **allow_multivalues** (``bool``) -- (optional, default: ``True``)
          If ``True`` multiple values for the same entity attribute will be
          put in the same SQL column, separated by ``multivalue_separator``.
          If ``False``, this function will raise an error
          should any attribute have a ``max`` greater than 1.
        * **multivalue_separator** (``str``) -- (optional, default: ``,``)
          "Character" separating multiple values for the same attribute
          in a given entity; used only if ``allow_multivalues`` is ``True``.
    """
    check_available()  # breaks if not
    path = url
    if _os.path.exists(path):
        raise ValueError(f"File '{path}' already exists")

    allow = options.get('allow_multivalues', False)
    allow = True  # TODO
    comma = options.get('multivalue_separator', ',')
    with open(path, 'w', newline='') as f:
        f.write(f"{_pre_everything()}\n")
        for entity in heimdall.getEntities(tree):
            eid = entity.get('id')
            attributes = heimdall.getAttributes(entity)
            items = heimdall.getItems(tree, lambda n: n.get('eid') == eid)
            f.write(f"\n{_pre_create_table(eid)}\n")
            f.write(f"{_create_table(tree, eid, attributes, allow)}\n")
            f.write(f"{_pre_dump_data(eid)}\n")
            f.write(f"{_dump_data(items, eid, attributes, comma)}\n")
            f.write(f"{_post_dump_data()}\n")


def _pre_everything():
    text = "-- SQL dump, from a HERA database, using pyHeimdall\n"
    text += "-- ------------------------------------------------------\n"
    text += f"-- pyHeimdall version	{heimdall.__version__}"
    return text


def _pre_create_table(eid, drop=True):
    comment = f"--\n-- Table structure for table `{eid}`\n--"
    drop_table = f"\nDROP TABLE IF EXISTS `{eid}`;" if drop else ""
    return comment + drop_table


def _create_table(tree, eid, attributes, allow_multivalues):
    text = f"CREATE TABLE `{eid}` (\n"
    for attribute in attributes:
        (aid, pid) = _get_identifiers(attribute)
        if aid is None and pid is None:
            continue  # this unused attribute won't be part of the dump
        _id = aid if aid is not None else pid
        max = attribute.get('max')
        max = int(max) if max else None
        if not allow_multivalues and (max is None or max > 1):
            fault = f"{eid}.{aid}.max={max}"
            raise ValueError(f"Repeatable attributes not supported ({fault})")
        min = int(attribute.get('min') or 0)
        min = "NOT NULL" if min > 0 else "DEFAULT NULL"
        type = _get_type(tree, attribute)
        text += f"`{_id}` {type} {min},\n"
    if text.endswith(',\n'):
        text = text[:-2] + '\n'  # remove last comma
    text += ") ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;"
    return text


def _get_identifiers(attribute):
    return (attribute.get('id'), attribute.get('pid'))


def _get_type(tree, attribute):
    type_node = heimdall.util.get_node(attribute, 'type')
    if type_node is not None:
        type_ = type_node.text
    else:
        pid = attribute.get('pid')
        property_ = heimdall.getProperty(tree, lambda n: n.get('id') == pid)
        type_node = heimdall.util.get_node(attribute, 'type')
        if type_node is not None:
            type_ = type_node.text
        else:
            type_ = 'text'
    if type_ == 'datetime':
        return 'date'
    if type_ == 'number':
        return 'int'  # TODO unsigned
    rules = heimdall.util.get_nodes(attribute, 'rule')
    length = 255
    for rule in rules:
        if rule.text.startswith(LENGTH_PREFIX):
            length = int(rule.text[len(LENGTH_PREFIX):])
    if length > 1:
        return f'varchar({length})'
    return 'char(1)'


def _pre_dump_data(eid, lock=True):
    comment = f"--\n-- Dumping data for table `{eid}`\n--"
    lock_table = f"\nLOCK TABLE `{eid}` WRITE;" if lock else ""
    return comment + lock_table


def _dump_data(items, eid, attributes, multivalues_separator):
    if len(items) == 0:
        return f"-- No data in table `{eid}`"
    text = f"INSERT INTO `{eid}` VALUES "
    dumps = list()
    for item in items:
        values = list()
        for attribute in attributes:
            values.append(_dump_value(item, attribute, multivalues_separator))
        dumps.append(f"({','.join(values)})")
    text += ','.join(dumps)
    text += ";"
    return text


def _dump_value(item, attribute, multivalues_separator):
    (aid, pid) = _get_identifiers(attribute)
    values = [f"'{v}'" for v in heimdall.getValues(item, pid)]  # TODO kwargs
    if len(values) < 1:
        return 'NULL'
    return multivalues_separator.join(values)


def _post_dump_data(lock=True):
    return "UNLOCK TABLES;"



# -*- coding: utf-8 -*-
from .heimdall import CONNECTORS_IN as _IN, CONNECTORS_OUT as _OUT
from functools import wraps


def connector_in(format):
    # just two lines to support syntax: @connector_in('format')
    # as long as syntax: @connector_in(['format1', 'format2'])
    if type(format) is not list:
        format = [format, ]

    def decorator(function):
        # register ``function`` as an IN connector
        for f in format:
            _IN[f] = function

        @wraps(function)
        def wrapper(*args, **kwargs):
            # call ``function`` and return its result
            return function(*args, **kwargs)
        return wrapper
    return decorator


def connector_out(format):
    # just two lines to support syntax: @connector_out('format')
    # as long as syntax: @connector_out(['format1', 'format2'])
    if type(format) is not list:
        format = [format, ]

    def decorator(function):
        # register ``function`` as an OUT connector
        for f in format:
            _OUT[f] = function

        @wraps(function)
        def wrapper(*args, **kwargs):
            # call ``function`` and return its result
            return function(*args, **kwargs)
        return wrapper
    return decorator



# -*- coding: utf-8 -*-

"""
Provides utility functions around HERA attributes.

:copyright: The pyHeimdall contributors.
:licence: Afero GPL, see LICENSE for more details.
:SPDX-License-Identifier: AGPL-3.0-or-later
"""
import heimdall
from . import get_node as _get_node, set_language as _set_language
from lxml import etree


'''
class Relationship:
    def __init__(self, eid, source, target):
        self.eid = eid  # a Relationship is an Entity, it's its id
        self.source = source  # source attribute pid
        self.target = target  # target attribute pid
'''


def refactor_relationship(tree, relationship, eid, euid, pid, cleanup=True):
    """TODO
    """
    R_ID = relationship['eid']
    R_SOURCE = relationship['source']
    R_TARGET = relationship['target']
    # TODO only create property if not exists
    p = heimdall.createProperty(tree, id=pid)
    e = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
    # TODO only create attribute if not exists
    a = heimdall.createAttribute(e, pid=pid)
    # iterate over all items belonging to the relationship entity
    items = heimdall.getItems(tree, lambda n: n.get('eid') == R_ID)
    for old in items:
        source = heimdall.getMetadata(old, R_SOURCE)[0].text
        target = heimdall.getMetadata(old, R_TARGET)[0].text

        def is_source(item):
            is_of_entity = item.get('eid') == eid
            has_unique_id = False
            # Unique id shouldn't be a repeatable attribute,
            # but we know what real world looks like. Thus,
            # try to not break in this edge case, and let's
            # hope our caller knows what she does.
            for v in heimdall.getValues(item, euid):
                has_unique_id = has_unique_id or (v == source)
            return is_of_entity and has_unique_id

        # get the item which must contain the new repeatable metadata
        now = heimdall.getItem(tree, is_source)
        etree.SubElement(now, 'metadata', pid=pid).text = target
        if cleanup:
            # delete `old` relationship item, because it is
            # now represented by new metadata in item `now`
            old.getparent().remove(old)
    if cleanup:
        # delete the `relationship` entity, as there are no more items using it
        heimdall.deleteEntity(tree, lambda n: n.get('id') == R_ID)


def merge_l10n_attributes(
        tree, eid, languages,
        pid=None, aid=None,
        cleanup=True, update_items=True
        ):
    """Merge attributes that are in fact translations of one another.

    :param tree: HERA elements tree
    :param eid: Identifier of the HERA entity in ``tree``
            containing all attributes to merge
    :param languages: dict containing attribute ids (aid) as keys
            and language codes as values
            it's not the other way around in case you have more than one
            attribute for the same language
    :param pid: HERA attribute pid of merged attribute
    :param aid: (optional, default ``{pid}_attr``): HERA attribute id
            (aid) other attributes must be merged into; if ``aid`` is not
            the id of an attribute already present in ``entity``, an attribute
            of this identifier and of type ``text`` will be created
    :param cleanup: (optional, default: ``True``) True if attributes in
            ``languages`` should be removed from ``entity`` after they
            are merged; ``base_aid`` is of course never removed.
    :param update_items: (optional, default: ``True``) True if items in
            ``tree`` should be updated with the new ``pid`` and ``aid``.

    Usage example: ::

      >>> import heimdall
      >>> tree == ...  # load HERA tree
      >>> # let's say `tree` contains this entity:
      >>> # <entity id='person'>
      >>> #     <attribute pid='name_de' id='person.name_de_attr'>
      >>> #         <name>Personenname</name>
      >>> #     <attribute>
      >>> #     <attribute pid='name_en' id='person.name_en_attr'>
      >>> #         <name>Name</name>
      >>> #         <description>Name of the person</description>
      >>> #     <attribute>
      >>> #     <attribute pid='name_fr' id='person.name_fr_attr'>
      >>> #         <name>Nom</name>
      >>> #         <description>Nom de la personne</description>
      >>> #     <attribute>
      >>> # </entity>
      >>> # MERGE "columns"/attributes that are translations of one another
      >>> # give ourselves a property for merging human-readable names
      >>> heimdall.createProperty(
      >>>     tree, 'dcmi:title', type='text',
      >>>     name={'en': "Title", 'fr': "Titre", },
      >>>     description={
      >>>         'en': "A name given to the resource.",
      >>>         'fr': "Nom de la ressource.",
      >>>         },
      >>>     uri=[
      >>>         'http://purl.org/dc/terms/title',
      >>>         'http://datacite.org/schema/kernel-4/title',
      >>>         ],
      >>>     )
      >>> # merge the_people names
      >>> e = heimdall.getEntity(tree, lambda n: n.get('id') == 'person')
      >>> merge_l10n_attributes(tree, e, {
      >>>     'person.name_de_attr': 'de',
      >>>     'person.name_en_attr': 'en',
      >>>     'person.name_fr_attr': 'fr',
      >>>     },
      >>>     aid='person.name',
      >>>     pid='dcmi:title')
      >>> # now entity person looks like that:
      >>> # <entity id='person'>
      >>> #     <attribute pid='dcmi:title' id='person.name'>
      >>> #         <name xml:lang='de'>Personenname</name>
      >>> #         <name xml:lang='en'>Name</name>
      >>> #         <name xml:lang='fr'>Nom</name>
      >>> #         <description xml:lang='en'>Name of the person</description>
      >>> #         <description xml:lang='fr'>Nom de la personne</description>
      >>> #     <attribute>
      >>> # </entity>
    """
    entity = heimdall.getEntity(tree, lambda n: n.get('id') == eid)
    if entity is None:
        raise ValueError(f"Entity '{eid}' doesn't exist")
    # Check all attr languages have same pid and type
    aids = list(k for k in languages.keys())
    first_aid = aids[0]
    first_a = heimdall.getAttribute(entity, lambda n: n.get('id') == first_aid)
    if first_a is None:
        raise ValueError(f"Unknown attribute identifier '{first_aid}'")
    base_pid = pid or first_a.get('pid')
    base_aid = aid or f'{base_pid}_attr'
    base_type = _get_node(first_a, 'type')
    if base_type is not None:
        base_type = base_type.text
    amin = 0
    # first pass: consistency checks
    for aid in aids[1:]:
        attr = heimdall.getAttribute(entity, lambda a: a.get('id') == aid)
        if attr is None:
            raise ValueError(f"Attribute '{aid}' doesn't exist in {eid}")
        atype = _get_node(attr, 'type')
        if atype is not None:
            atype = atype.text
        if atype != base_type:
            raise ValueError("Attributes don't all have the same type")
        amin = max(amin, int(attr.get('min')))
    # get or create the attribute that will merge all others of `languages`
    base_a = heimdall.getAttribute(entity, lambda n: n.get('id') == base_aid)
    if base_a is None:
        base_a = heimdall.createAttribute(
                entity, id=base_aid, pid=base_pid,
                min=amin, max=None,
                type=base_type,
                )
    # second pass: merge attributes, and ...
    # * ... if `cleanup` delete `language` attributes (except base_atr ofc)
    # * ... if `update_items`, update_items to use id = base_a.id / base_aid
    items = []
    if update_items:
        items = heimdall.getItems(tree, lambda n: n.get('eid') == eid)
    for aid, language in languages.items():
        attr = heimdall.getAttribute(entity, lambda a: a.get('id') == aid)
        node = _get_node(attr, 'name')
        if node is not None:
            _set_language(node, language)
            base_a.append(node)
        node = _get_node(attr, 'description')
        if node is not None:
            _set_language(node, language)
            base_a.append(node)

        if cleanup and (aid != base_aid):
            attr.getparent().remove(attr)
        for item in items:
            for metadata in heimdall.getMetadata(item, aid=aid):
                metadata.set('aid', base_aid)
                metadata.set('pid', base_pid)
                _set_language(metadata, language)


__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'



# -*- coding: utf-8 -*-

from .tree import (
    get_node, get_nodes, get_root,
    create_node, create_nodes,
    get_language, set_language,
    update_node_value, maybe_update_node_values,
    update_node_children, maybe_update_node_children,
    )
from .entities import update_entities
from .attributes import merge_l10n_attributes, refactor_relationship
from .properties import delete_unused_properties, merge_properties

__all__ = [
    'get_node', 'get_nodes', 'get_root',
    'create_node', 'create_nodes',
    'get_language', 'set_language',
    'update_node_value', 'maybe_update_node_values',
    'update_node_children', 'maybe_update_node_children',

    'update_entities',
    'merge_l10n_attributes',
    'refactor_relationship',
    'delete_unused_properties', 'merge_properties',

    '__copyright__', '__license__',
    ]
__copyright__ = "Copyright the pyHeimdall contributors."
__license__ = 'AGPL-3.0-or-later'



# -*- coding: utf-8 -*-
import csv as _csv
import os as _os
import heimdall
from ..decorators import connector_in, connector_out
from glob import glob as _glob


@connector_in('csv')
def getDatabase(url, **options):
    r"""Imports a database from one or more CSV files

    :param \**options: Keyword arguments, see below.
    :return: HERA element tree
    :rtype: lxml.ElementTree

    :Keyword arguments:
        * **url** (``str``) -- Pattern of CSV files to read from

    .. ERROR::
       This feature is not implemented yet.

       Interested readers can submit a request to further this topic.
       See the ``CONTRIBUTING`` file at the root of the repository for details.
    """
    tree = heimdall.createDatabase()
    for path in _glob(url):
        _file2hera(tree, path, **options)
    return tree


def _file2hera(tree, url, **options):
    newline = ''
    header = options.get('header', True)
    comma = options.get('delimiter', ',')
    quote = options.get('quotechar', '"')
    pipe = options.get('multivalue', '|')
    eid = options.get('eid', _os.path.splitext(_os.path.basename(url))[0])
    with open(url, newline=newline) as f:
        reader = _csv.DictReader(f, delimiter=comma, quotechar=quote)
        columns = None
        for row in reader:
            if header and columns is None:
                columns = list(row.keys())
                _row2properties(tree, columns)
                _row2entity(tree, eid, columns)
            _row2item(tree, row)
    return tree


def _row2properties(tree, header):
    properties = list()
    for column in header:
        properties.append(heimdall.createProperty(tree, id=column))
    return properties


def _row2entity(tree, eid, header):
    entity = heimdall.createEntity(tree, id=eid)
    for column in header:
        heimdall.createAttribute(entity, pid=column, id=column)
    return entity


def _row2item(tree, row):
    heimdall.createItem(tree, **row)


@connector_out('csv')
def serialize(tree, url, **options):
    r"""Serializes a HERA elements tree into CSV files

    :param tree: HERA elements tree
    :param url: Path to an existing directory
    :param \**options: (optional) Keyword arguments, see below.
    :Keyword arguments:
        * **header** (``bool``) -- (optional, default: ``True``)
          If ``True``, first line in each CSV is a header containing column
          names; if ``False``, first line will be the first item in ``tree``
        * **delimiter** (``str``) -- (optional, default: ``,``)
          CSV column delimiter
        * **quotechar** (``str``) -- (optional, default: ``"``)
          CSV "quoting character"; quotes will only be used if necessary
        * **multivalue** (``str``) -- (optional, default: ``|``)
          CSV multivalue delimiter

    This function can be used to export an HERA elements tree as CSV files.
    One CSV file is created per entity in ``tree``; this CSV has one column
    per attribute in this entity, and one line (excluding header if any)
    per item belonging to this entity in ``tree``.

    Each CSV file path will be ``<url>/<eid>.csv``, with ``eid`` being each
    entity's identifier in ``tree``.
    If a given entity doesn't have any attribute, the file will be empty.
    If no item belongs to this entity in ``tree``, the file will be empty,
    bar the header if any.
    """
    folder = url
    if not _os.path.isdir(folder):
        raise ValueError("Option 'path' must be a directory")

    header = options.get('header', True)
    comma = options.get('delimiter', ',')
    quote = options.get('quotechar', '"')
    pipe = options.get('multivalue', '|')

    for entity in heimdall.getEntities(tree):
        eid = entity.get('id')
        path = _os.path.join(folder, f'{eid}.csv')
        attributes = heimdall.getAttributes(entity)
        header = [_attr2col(a) for a in attributes]
        with open(path, 'w', newline='') as f:
            writer = _csv.writer(
                f,
                delimiter=comma,
                quotechar=quote,
                quoting=_csv.QUOTE_MINIMAL)
            if header:
                writer.writerow(header)
            items = heimdall.getItems(tree, lambda n: n.get('eid') == eid)
            for item in items:
                row = _item2row(item, attributes, pipe)
                writer.writerow(row)


def _attr2col(attribute):
    """Deduces column name from `attribute`

    :param attribute: HERA attribute element
    """
    name = attribute.get('id')
    if name is not None:
        return name
    name = attribute.get('pid')
    entity = attribute.getparent()
    if name is not None:
        eid = f"{entity.get('id')}." if entity is not None else ''
        return f'{eid}{name}'
    # NOTE: from here, an error could be raised IF we take into
    #       account that a pid is mandatory for each attribute
    #       and that this feature should only be used as part
    #       of the serialization process of valid HERA trees
    if entity is not None:
        attributes = heimdall.getAttributes(entity)
        index = attributes.index(attribute)
        return f"{entity.get('id')}.{index}"
    return "?"


def _item2row(item, attributes, pipe):
    """TODO

    :param item: HERA item element
    :param attributes: Attribute identifiers
    :param pipe: Multivalue separator
    """
    row = list()
    for a in attributes:
        aid = a.get('id')
        pid = a.get('pid')
        metadata = heimdall.getMetadata(item, aid=aid, pid=pid)
        value = pipe.join((m.text or '') for m in metadata)
        row.append(value)
    return row
